#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/stat.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <stdint.h>

#define SOCK_PATH "/tmp/cexec.sock"
#define LISTEN_BACKLOG 4
#define BUF_SIZE 16

#define handle_error(msg) do { perror(msg); exit(EXIT_FAILURE); } while(0)

void monitor(int cfd) {
    char buf[BUF_SIZE];
    ssize_t nread;
    struct msghdr message;
    struct iovec iov;

    memset(&message, 0, sizeof(struct msghdr));
    message.msg_name = NULL;
    message.msg_namelen = 0;
    message.msg_iov = &iov;
    message.msg_iovlen = 1;
    message.msg_control = NULL;
    message.msg_controllen = 0;

    iov.iov_base = buf;
    iov.iov_len = BUF_SIZE;
    nread = recvmsg(cfd, &message, 0);
    if (nread == -1)
        handle_error("recvmsg hello");
    printf("%s\n", buf);

    iov.iov_base = buf;
    iov.iov_len = BUF_SIZE;

    /* set up cmsg for file descriptor. We will receive remote stdout */
    struct cmsghdr *cmsg;
    char cmsgbuf[CMSG_SPACE(sizeof(int))] = {0};
    message.msg_control = cmsgbuf;
    message.msg_controllen = sizeof(cmsgbuf);
    cmsg = CMSG_FIRSTHDR(&message);
    int *fdptr = (int *)CMSG_DATA(cmsg);
    nread = recvmsg(cfd, &message, 0);
    if (nread == -1)
        handle_error("recvmsg fd");
    printf("remote fd number %d\n", ((int *)buf)[0]);
    printf("new fd number %d\n", *fdptr);
    
    /* this message should print on remote stdout */
    dup2(*fdptr, 1);
    printf("printing to new stdout\n");
    exit(0);
}

int main() {
    int sfd, cfd;
    struct sockaddr_un my_addr, peer_addr;
    socklen_t peer_addr_size;
    pid_t pid;

    sfd = socket(AF_UNIX, SOCK_SEQPACKET, 0);
    if (sfd == -1)
        handle_error("socket");

    unlink(SOCK_PATH); /* this may fail but does not hurt */

    /* Clear structure */
    memset(&my_addr, 0, sizeof(struct sockaddr_un));
    my_addr.sun_family = AF_UNIX;
    strncpy(my_addr.sun_path, SOCK_PATH, sizeof(my_addr.sun_path) - 1);

    /* set umask to 0 to allow world-writable socket */
    mode_t mask = umask(0);
    if (bind(sfd, (struct sockaddr *) &my_addr,
            sizeof(struct sockaddr_un)) == -1)
        handle_error("bind");
    umask(mask);

    if (listen(sfd, LISTEN_BACKLOG) == -1)
        handle_error("listen");

    peer_addr_size = sizeof(struct sockaddr_un);
    while ((cfd = accept(sfd, (struct sockaddr *) &peer_addr,
                    &peer_addr_size)) != -1) {
        if ((pid = fork()) == -1)
            handle_error("fork");
        if (!pid) {
            close(sfd);
            monitor(cfd);
        }
        close(cfd);
    }
    handle_error("accept");
    return 0;
}

